/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2.3)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2008 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: xi_ari.hpp,v 1.4 2008/01/23 08:39:21 cxsc Exp $ */

//============================================================================
//
//                              Program/Module
//                                   from
//                 C++ TOOLBOX FOR VERIFIED COMPUTING I
//                         Basic Numerical Problems
//
//      Copyright (c) 1995   Rolf Hammer, Matthias Hocks, Dietmar Ratz
//
// This program/module is free software for non-commercial use. For details
// on theory, algorithms, and programs, see the book
//
//  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  C++ Toolbox for
//  Verified Computing I - Basic Numerical Problems. Springer-Verlag,
//  Heidelberg, New York, 1995.
//
// This program/module is distributed WITHOUT ANY WARRANTY. For details,
// see the "Disclaimer / Legal Matters" of the book (page iv).
//
//============================================================================
//----------------------------------------------------------------------------
// File: lxi_ari (header)
// Purpose: Definition of an extended interval arithmetic which allows the
//    division by an interval containing zero, based on the extended 
//    staggered interval arithnetic; class lx_interval:
// Global type:
//    KindType         : component type of extended intervals
// Global function:
//    EmptyIntval()    : empty set as irregular interval
// Class xinterval:
//    operators %, -, &: operators of extended interval arithmetic
//    operator =       : assignment operator
// Updates:
//    04.03.1996 : Modification of extended interval operations and kind type
//----------------------------------------------------------------------------
#ifndef __LXI_ARI_HPP
#define __LXI_ARI_HPP

#include "lx_interval.hpp"     // Extended staggered interval arithmetic
#include "lx_ivector.hpp"      // Extended interval vector arithmetic

using namespace cxsc;
using namespace std;

typedef enum { Finite, PlusInfty, MinusInfty, Double, Empty } KindType;

extern interval EmptyIntval ( );   // Irregular (empty) interval

class xlx_interval {               // Extended intervals according
  private:                         // to the above definition
    KindType  kind;                //-----------------------------
    lx_real   inf, sup;

  public:
    xlx_interval ( );
    xlx_interval ( const KindType&, const lx_real&, const lx_real& );
    xlx_interval ( const xlx_interval& );

    xlx_interval& operator= ( const xlx_interval& );

    friend xlx_interval operator- ( const lx_real&, const xlx_interval& );
    friend xlx_interval operator% ( const lx_interval&, const lx_interval& );
    friend lx_ivector   operator& ( const lx_interval&, const xlx_interval& );
};

xlx_interval operator- ( const lx_real&,     const xlx_interval& );
xlx_interval operator% ( const lx_interval&, const lx_interval& );
lx_ivector   operator& ( const lx_interval&, const xlx_interval& );
#endif

