/*
**  Copyright (C) 1999-2006 F. Blomquist, M. Braeuer,
**                          W. Hofschuster, W. Kraemer
**                          Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//////////////////////////////////////////////////////////////
//
//       Implementation of class lx_itaylor in lxitaylor.cpp
//
//////////////////////////////////////////////////////////////

#include "lxitaylor.hpp"

///////////////////////////////////////////////////////////////
//
//                     class lx_itaylor
//
///////////////////////////////////////////////////////////////

namespace taylor {

lx_ivector lx_itaylor::faks(1);
int lx_itaylor::initialized=0;

void lx_itaylor::initialize()
{
 Resize(lx_itaylor::faks,0,170);
 lx_itaylor::faks[0]=lx_interval(0,l_interval(1.0));
 lx_itaylor::faks[1]=lx_interval(0,l_interval(1.0));

 for(int i=2; i<=170; i++)
   lx_itaylor::faks[i]=lx_itaylor::faks[i-1]*lx_interval(0,l_interval(i));
}

//-----------------------------------------------------------------------

// Constructors:

lx_itaylor::lx_itaylor()
{
 if (!lx_itaylor::initialized)
 	{lx_itaylor::initialize();lx_itaylor::initialized=1;};
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(const lx_itaylor& s)
{
 if (!lx_itaylor::initialized)
 	{lx_itaylor::initialize(); lx_itaylor::initialized=1;};
 p=s.p;
 Resize(tayl,0,p);
 tayl=s.tayl;
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(int order)
{
	if (!lx_itaylor::initialized)
		{lx_itaylor::initialize();lx_itaylor::initialized=1;};
	if(order<0 || order>170)
		{
    		std::cerr << "lx_itaylor::lx_itaylor: incorrect order! 0<=order<=170"
              	 	 << std::endl;
    		exit(1);
		}
	p=order;
 	Resize(tayl,0,p);
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(int order, const real& value)
{
	if (!lx_itaylor::initialized)
		{lx_itaylor::initialize();lx_itaylor::initialized=1;};
 	if (order<0 || order>170)
  	{
    	std::cerr << "lx_itaylor::lx_itaylor: incorrect order! 0<=order<=170"
              	 << std::endl;
    	exit(1);
  	}
 	p=order;
 	Resize(tayl,0,p);
 	tayl[0] = lx_interval(0,l_interval(value));
   if(p>0)
   {
     	tayl[1]=interval(1.0);
     	for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
   }
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(int order, const l_real& value)
{
	if (!lx_itaylor::initialized)
		{lx_itaylor::initialize();lx_itaylor::initialized=1;};
 	if(order<0 || order>170)
  	{
    	std::cerr << "lx_itaylor::lx_itaylor:  incorrect order! 0<=order<=170" 
              	 << std::endl;
    	exit(1);
  	}
 	p=order;
 	Resize(tayl,0,p);
 	tayl[0] = lx_interval(0,value);
   if(p>0)
   {
     	tayl[1]=interval(1.0);
     	for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
   }
}

lx_itaylor::lx_itaylor(int order, const lx_real& value)
{
	if (!lx_itaylor::initialized)
	{lx_itaylor::initialize();lx_itaylor::initialized=1;};
	if(order<0 || order>170)
	{
		std::cerr << "lx_itaylor::lx_itaylor:  incorrect order! 0<=order<=170" 
				<< std::endl;
		exit(1);
	}
	p=order;
	Resize(tayl,0,p);
	tayl[0] = lx_interval(value);
	if(p>0)
	{
		tayl[1]=interval(1.0);
		for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
	}
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(int order, const interval& value)
{
	if (!lx_itaylor::initialized)
	{lx_itaylor::initialize();lx_itaylor::initialized=1;};
 	if(order<0 || order>170) 
  	{
    	std::cerr << "lx_itaylor::lx_itaylor: incorrect order! 0<=order<=170"
                << std::endl;
    	exit(1);
  	}
 	p=order;
 	Resize(tayl,0,p);
 	tayl[0]=lx_interval(value);
 	if(p>0)
   { 
     	tayl[1]=interval(1.0);
     	for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
   }
}

//-----------------------------------------------------------------------

lx_itaylor::lx_itaylor(int order, const l_interval& value)
{
	if (!lx_itaylor::initialized)
		{lx_itaylor::initialize();lx_itaylor::initialized=1;};
 	if(order<0 || order>170) 
  	{
    	std::cerr << "lx_itaylor::lx_itaylor: incorrect order! 0<=order<=170" 
                << std::endl;
    	exit(1);
  	}
 	p=order;
 	Resize(tayl,0,p);
 	tayl[0]=value;
 	if(p>0)
   { 
     	tayl[1]=interval(1.0);
     	for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
   }
}

lx_itaylor::lx_itaylor(int order, const lx_interval& value)
{
	if (!lx_itaylor::initialized)
	{lx_itaylor::initialize();lx_itaylor::initialized=1;};
	if(order<0 || order>170) 
	{
		std::cerr << "lx_itaylor::lx_itaylor: incorrect order! 0<=order<=170" 
				<< std::endl;
		exit(1);
	}
	p=order;
	Resize(tayl,0,p);
	tayl[0]=value;
	if(p>0)
	{ 
		tayl[1]=interval(1.0);
		for(int i=2; i<=Ub(tayl);i++) tayl[i]=interval(0.0);
	}
}

// Functions for initialization of independent variables:
lx_itaylor var_lx_itaylor(int ord, const real& x)
{
 	lx_itaylor erg(ord,x);
 	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor var_lx_itaylor(int ord, const l_real& x)
{
 	lx_itaylor erg(ord,x);
 	return erg;
}

lx_itaylor var_lx_itaylor(int ord, const lx_real& x)
{
	lx_itaylor erg(ord,x);
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor var_lx_itaylor(int ord, const interval& x)
{
 	lx_itaylor erg(ord,x);
 	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor var_lx_itaylor(int ord, const l_interval& x)
{
 	lx_itaylor erg(ord,x);
 	return erg;
}

lx_itaylor var_lx_itaylor(int ord, const lx_interval& x)
{
	lx_itaylor erg(ord,x);
	return erg;
}

//-----------------------------------------------------------------------

// Functions for initialization of constants:
lx_itaylor const_lx_itaylor(int ord, const real& c)
{
 	lx_itaylor erg(ord);
 	erg.tayl[0]=lx_interval(0,l_interval(c));
 	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
 	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor const_lx_itaylor(int ord, const l_real& c)
{
 	lx_itaylor erg(ord);
 	erg.tayl[0]=lx_interval(0,c);
 	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
 	return erg;
}

lx_itaylor const_lx_itaylor(int ord, const lx_real& c)
{
	lx_itaylor erg(ord);
	erg.tayl[0]=lx_interval(c);
	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor const_lx_itaylor(int ord, const interval& c)
{
 	lx_itaylor erg(ord);
 	erg.tayl[0]=lx_interval(c);
 	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
 	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor const_lx_itaylor(int ord, const l_interval& c)
{
 	lx_itaylor erg(ord);
 	erg.tayl[0]=lx_interval(0,c);
 	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
 	return erg;
}

lx_itaylor const_lx_itaylor(int ord, const lx_interval& c)
{
	lx_itaylor erg(ord);
	erg.tayl[0]=c;
	for(int i=1; i<=Ub(erg.tayl);i++) erg.tayl[i]=interval(0.0);
	return erg;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

// assignment operators:

lx_itaylor lx_itaylor::operator=(const lx_itaylor& s)
{
    p=s.p;  // p: order of the left part; s.p: order of the right part; 
    Resize(tayl,0,s.p);
    tayl=s.tayl;
    return *this; // The left operand gets the order of the right operand!
}

//-----------------------------------------------------------------------

lx_itaylor lx_itaylor::operator=(int c)
{
    tayl[0] = lx_interval(0,l_interval(c));
    for (int j=1; j<=p; j++) tayl[j]=0.0;
    return *this;
}

//-----------------------------------------------------------------------

lx_itaylor lx_itaylor::operator=(const real& c)
{
    tayl[0] = lx_interval(0,l_interval(c));
    for (int j=1; j<=p; j++) tayl[j]=0.0;
    return *this;
}

//-----------------------------------------------------------------------

lx_itaylor lx_itaylor::operator=(const l_real& c)
{
    tayl[0] = c;
    for (int j=1; j<=p; j++) tayl[j]=0.0;
    return *this;
}

lx_itaylor lx_itaylor::operator=(const lx_real& c)
{
	tayl[0] = c;
	for (int j=1; j<=p; j++) tayl[j]=0.0;
	return *this;
}

//-----------------------------------------------------------------------

lx_itaylor lx_itaylor::operator=(const interval& c)
{
    tayl[0] = lx_interval(c);
    for (int j=1; j<=p; j++) tayl[j]=0.0;
    return *this;
}

//-----------------------------------------------------------------------

lx_itaylor lx_itaylor::operator=(const l_interval& c)
{
    tayl[0] = lx_interval(0,c);
    for (int j=1; j<=p; j++) tayl[j]=0.0;
    return *this;
}

lx_itaylor lx_itaylor::operator=(const lx_interval& c)
{
	tayl[0] = c;
	for (int j=1; j<=p; j++) tayl[j]=0.0;
	return *this;
}
//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

// class components:

// returning the maximal order
int get_order(const lx_itaylor& x)
{
 	return x.p;
}

//-----------------------------------------------------------------------

// returning all Taylor-coefficients by an lx_interval vector
lx_ivector get_all_coef(const lx_itaylor& x)
{
 	return x.tayl;
}

//-----------------------------------------------------------------------

// returning Taylor-coefficient of order j
lx_interval get_j_coef(const lx_itaylor& x, int j)
{
	 return x.tayl[j];
}

//-----------------------------------------------------------------------

// returning derivative of order j,  j <= 170;
lx_interval get_j_derivative(const lx_itaylor& x, int j)
{
	return x.tayl[j]*lx_itaylor::faks[j];
}

//-----------------------------------------------------------------------

// Output of all taylor coefficients:
void print_base10_lx_itaylor(const lx_itaylor& x)
// Output of the taylor coefficients using base 10,
// i.e. additional logarithmic calculations are necessary,
// which can reduce the accuracy of the output, but not the
// actual internal accuracy of the taylor cofficients.
{
	std::cerr <<"Output: lx_itaylor of order " << x.p << " " << std::endl;
	for(int i=Lb(x.tayl); i<=Ub(x.tayl);i++) 
  	{
   	std::cerr << "i = " << i << "  component: " << x.tayl[i] << std::endl;
  	};
 	std::cerr << std::endl;
}

void print_base2_lx_itaylor(const lx_itaylor& x)
// Output of the taylor coefficients using base 2,
// i.e. the maximum internal accuracy of the coefficients
// is printed.
{
	int stp;
	string s;
	std::cerr <<"Output: lx_itaylor of order " << x.p << " " << std::endl;
	std::cout << SaveOpt;
	for(int i=Lb(x.tayl); i<=Ub(x.tayl);i++) 
	{
		s = "";
		stp = StagPrec(x.tayl[i]);
		std::cout << SetDotPrecision(16*stp,16*stp) << Scientific;
		s << x.tayl[i];
		std::cerr << "i = " << i << "  component: " << s << std::endl;
	};
	std::cout << RestoreOpt;
	std::cerr << std::endl;
}

//-----------------------------------------------------------------------

// Overloading of operators:

//-----------------------------------------------------------------------

// - operator:
lx_itaylor operator-(const lx_itaylor& x)
{
	int order=get_order(x);
	lx_itaylor erg(order);
	for(int j=Lb(x.tayl); j<=Ub(x.tayl); j++) erg.tayl[j]= -x.tayl[j];
 	return erg;
}

//-----------------------------------------------------------------------

// Operators with two operands:  +,-,*,/  for (lx_itaylor, lx_itaylor):
// All operands are independent variables.
lx_itaylor operator-(const lx_itaylor& x, const lx_itaylor& y)
{
	int order1=get_order(x);
 	int order2=get_order(y);
 	if(order1 != order2) 
  	{
   	std::cerr << "Error in lx_itaylor, operator - : different orders " 
             	 << std::endl;
   	exit(1);
  	};

 	lx_itaylor erg(order1);
 	for(int j=Lb(x.tayl); j<=Ub(x.tayl); j++) erg.tayl[j]= x.tayl[j]-y.tayl[j];
 	return erg; 
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const lx_itaylor& y)
{
 int order1=get_order(x);
 int order2=get_order(y);
 if(order1 != order2) 
  {
   	std::cerr << "Error in lx_itaylor, operator + : different orders " 
             << std::endl;
   	exit(1);
  };

 lx_itaylor erg(order1);
 for(int j=Lb(x.tayl); j<=Ub(x.tayl); j++) erg.tayl[j]= x.tayl[j]+y.tayl[j];
 return erg; 
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const lx_itaylor& y)
{
	int order1=get_order(x);
 	int order2=get_order(y);
 	if (order1 != order2) 
  	{
   	std::cerr << "Error in lx_itaylor, operator * : different orders " 
                << std::endl;
   	exit(1);
  	};

 	lx_itaylor erg(order1);
 	lx_interval sum; 

	for(int j=0; j<=order1; j++) 
	{
  		sum = lx_interval(0,l_interval(0));
  		for(int i=0; i<=j; i++)
    		sum = sum + x.tayl[i]*y.tayl[j-i];
  		erg.tayl[j]= sum;
 	}
 	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const lx_itaylor& y)
{
	int order1(get_order(x));
 	int order2(get_order(y));
 	if (order1 != order2) 
  	{
   	std::cerr << "Error in lx_itaylor, operator / : different orders " 
                << std::endl;
   	exit(1);
  	};
 
 	if (0 <= y.tayl[0]) 
  	{
   	std::cerr << "Error in lx_itaylor, operator / : 0 in interval " 
				    << std::endl;
   	exit(1);
  	};

 	lx_itaylor erg(order1);
 	lx_interval sum;

 	for (int j=0; j<=order1; j++) 
 	{
  		sum = lx_interval(0,l_interval(0));
  		for (int i=1; i<=j; i++)
      {
    		sum = sum + y.tayl[i]*erg.tayl[j-i];
   	}
	   erg.tayl[j]= (x.tayl[j]-sum)/y.tayl[0];
 	}
 	return erg;
}

//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (interval, lx_itaylor):
// The operand of type interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = -y;
    erg.tayl[0] = x - y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = y;
    erg.tayl[0] = x + y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const interval& x, const lx_itaylor& y)
{
    if (0<=y.tayl[0])
    {
	 	std::cerr << "Error in lx_itaylor, operator / : 0 in interval " 
                << std::endl;
		exit(1);
    }; 
    lx_interval sum;
    int order(get_order(y));
    lx_itaylor w(order);
    w.tayl[0] = x / y.tayl[0];
    for (int k=1; k<=order; k++)
    {
	 	sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++) 
	   	sum = sum + y.tayl[j]*w.tayl[k-j];
	   w.tayl[k] = -sum / y.tayl[0];
    }
    return w;
}


// Operators with two operands:   +,-,*,/  for (l_interval, lx_itaylor):
// The operand of type interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const l_interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = -y;
    erg.tayl[0] = x - y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const l_interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = y;
    erg.tayl[0] = x + y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const l_interval& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const l_interval& x, const lx_itaylor& y)
{
    if (0<=y.tayl[0])
    {
		std::cerr << "Error in lx_itaylor, operator / : 0 in interval " 
                << std::endl;
		exit(1);
    }; 
    lx_interval sum;
    int order(get_order(y));
    lx_itaylor w(order);
    w.tayl[0] = x / y.tayl[0];
    for (int k=1; k<=order; k++)
    {
	 	sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++) 
	   	sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
    }
    return w;
}

// Operators with two operands:   +,-,*,/  for (lx_interval, lx_itaylor):
// The operand of type lx_interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_interval& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	erg = -y;
	erg.tayl[0] = x - y.tayl[0];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_interval& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	erg = y;
	erg.tayl[0] = x + y.tayl[0];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_interval& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_interval& x, const lx_itaylor& y)
{
	if (0<=y.tayl[0])
	{
		std::cerr << "Error in lx_itaylor, operator / : 0 in interval " 
				<< std::endl;
		exit(1);
	}; 
	lx_interval sum;
	int order(get_order(y));
	lx_itaylor w(order);
	w.tayl[0] = x / y.tayl[0];
	for (int k=1; k<=order; k++)
	{
		sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++) 
			sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
	}
	return w;
}

//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (l_itaylor, interval):
// The operand of type interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] - y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] + y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
    return erg;
}

// Operators with two operands:   +,-,*,/  for (lx_itaylor, l_interval):
// The operand of type interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const l_interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] - y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const l_interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] + y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const l_interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const l_interval& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
    return erg;
}

// Operators with two operands:   +,-,*,/  for (lx_itaylor, lx_interval):
// The operand of type lx_interval is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const lx_interval& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	erg = x;
	erg.tayl[0] = x.tayl[0] - y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const lx_interval& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	erg = x;
	erg.tayl[0] = x.tayl[0] + y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const lx_interval& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const lx_interval& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
	return erg;
}


//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (real, lx_itaylor):
// The operand of type real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = -y;
    erg.tayl[0] = x - y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = y;
    erg.tayl[0] = x + y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/( const real& x, const lx_itaylor& y)
{
    if (0<=y.tayl[0])
    {
	 	std::cerr << "Error in lx_itaylor, operator / : 0 in interval " 
                << std::endl;
		exit(1);
    }; 
    lx_interval sum;
    int order(get_order(y));
    lx_itaylor w(order);
    w.tayl[0] = x / y.tayl[0];
    for (int k=1; k<=order; k++)
    {
	 	sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++) 
	    	sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
    }
    return w;
}

// Operators with two operands:   +,-,*,/  for (int, lx_itaylor):
// The operand of type real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(int x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = -y;
    erg.tayl[0] = interval(x) - y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(int x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = y;
    erg.tayl[0] = interval(x) + y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(int x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) 
		 erg.tayl[j] = interval(x)*y.tayl[j];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(int x, const lx_itaylor& y)
{
    if (0<=y.tayl[0])
    {
		std::cerr << "Error in lx_itaylor, operator / : 0 in interval "
                << std::endl;
		exit(1);
    };
    lx_interval sum;
    int order(get_order(y));
    lx_itaylor w(order);
    w.tayl[0] = interval(x) / y.tayl[0];
    for (int k=1; k<=order; k++)
    {
	 	sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++)
	    	sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
    }
    return w;
}

//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (l_real, lx_itaylor):
// The operand of type l_real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const l_real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = -y;
    erg.tayl[0] = x - y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const l_real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    erg = y;
    erg.tayl[0] = x + y.tayl[0];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const l_real& x, const lx_itaylor& y)
{
    int order(get_order(y));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const l_real& x, const lx_itaylor& y)
{
    if (0<=y.tayl[0])
    {
		std::cerr << "Error in lx_itaylor, operator / : 0 in interval "
                << std::endl;
		exit(1);
    };
    lx_interval sum;
    int order(get_order(y));
    lx_itaylor w(order);
    w.tayl[0] = x / y.tayl[0];
    for (int k=1; k<=order; k++)
    {
		sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++)
	    	sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
    }
    return w;
}

// Operators with two operands:   +,-,*,/  for (lx_real, lx_itaylor):
// The operand of type lx_real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_real& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	erg = -y;
	erg.tayl[0] = x - y.tayl[0];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_real& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	erg = y;
	erg.tayl[0] = x + y.tayl[0];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_real& x, const lx_itaylor& y)
{
	int order(get_order(y));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x*y.tayl[j];
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_real& x, const lx_itaylor& y)
{
	if (0<=y.tayl[0])
	{
		std::cerr << "Error in lx_itaylor, operator / : 0 in interval "
				<< std::endl;
		exit(1);
	};
	lx_interval sum;
	int order(get_order(y));
	lx_itaylor w(order);
	w.tayl[0] = x / y.tayl[0];
	for (int k=1; k<=order; k++)
	{
		sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=k; j++)
			sum = sum + y.tayl[j]*w.tayl[k-j];
		w.tayl[k] = -sum / y.tayl[0];
	}
	return w;
}

//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (lx_itaylor, real):
// The operand of type real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] - y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] + y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const real& y)
{
    if (y==0)
    {
		std::cerr << "Error in lx_itaylor: division by 0"
                << std::endl;
		exit(1);
    };
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
    return erg;
}

// Operators with two operands:   +,-,*,/  for (lx_itaylor, int):
// The operand of type int is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, int y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] - interval(y);
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, int y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] + interval(y);
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, int y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*interval(y);
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, int y)
{
    if (y==0)
    {
		std::cerr << "Error in lx_itaylor: division by 0"
                << std::endl;
		exit(1);
    };
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/interval(y);
    return erg;
}


//-----------------------------------------------------------------------

// Operators with two operands:   +,-,*,/  for (lx_itaylor, l_real):
// The operand of type l_real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const l_real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] - y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const l_real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    erg = x;
    erg.tayl[0] = x.tayl[0] + y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const l_real& y)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
    return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const l_real& y)
{
    if (y==0)
    {
		std::cerr << "Error in lx_itaylor: division by 0"
                << std::endl;
		exit(1);
    };
    int order(get_order(x));
    lx_itaylor erg(order);
    for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
    return erg;
}


// Operators with two operands:   +,-,*,/  for (lx_itaylor, lx_real):
// The operand of type lx_real is assumed to be a constant and not
// an independent variable!

lx_itaylor operator-(const lx_itaylor& x, const lx_real& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	erg = x;
	erg.tayl[0] = x.tayl[0] - y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator+(const lx_itaylor& x, const lx_real& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	erg = x;
	erg.tayl[0] = x.tayl[0] + y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator*(const lx_itaylor& x, const lx_real& y)
{
	int order(get_order(x));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]*y;
	return erg;
}

//-----------------------------------------------------------------------

lx_itaylor operator/(const lx_itaylor& x, const lx_real& y)
{
	if (y==0)
	{
		std::cerr << "Error in lx_itaylor: division by 0"
				<< std::endl;
		exit(1);
	};
	int order(get_order(x));
	lx_itaylor erg(order);
	for (int j=0; j<=order; j++) erg.tayl[j] = x.tayl[j]/y;
	return erg;
}

//-----------------------------------------------------------------------

// Overloading the standard functions:

//-----------------------------------------------------------------------

// sqr-function
lx_itaylor sqr(const lx_itaylor& x)
{
    lx_interval sum;
    int order(get_order(x)),m;
    lx_itaylor erg(order);

    erg.tayl[0] = sqr(x.tayl[0]);
    for (int k=1; k<=order; k++)
    {
		m = (k+1) / 2;
		sum = lx_interval(0,l_interval(0));
		for (int j=0; j<=m-1; j++)
	    	sum = sum + x.tayl[j]*x.tayl[k-j];
		times2pown(sum,1);  // Multiplication with 2
		erg.tayl[k] = sum;
		if (k%2==0) 
			erg.tayl[k] += sqr(x.tayl[m]); // k even
    }
    return erg;
}

//-----------------------------------------------------------------------

// Square-root

lx_itaylor sqrt(const lx_itaylor& x)
{
    lx_interval sum,h;
    int order(get_order(x)),m;
    lx_itaylor erg(order);
    if (0<=x.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, sqrt: 0 in interval" << std::endl;
		exit(1);
    };
    erg.tayl[0] = sqrt(x.tayl[0]);
    h = erg.tayl[0];
    times2pown(h,1);
    for (int k=1; k<=order; k++)
    {
		m = (k+1) / 2;
		sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=m-1; j++) 
	    	sum = sum + erg.tayl[j]*erg.tayl[k-j];
		times2pown(sum,1);  // Multiplication with 2
		erg.tayl[k] = sum;
		if (k%2==0) 
			erg.tayl[k] += sqr(erg.tayl[m]); // k even 
		erg.tayl[k] = (x.tayl[k]-erg.tayl[k]) / h;
    }
    return erg;
}

//-----------------------------------------------------------------------

// sqrt(x,n)

lx_itaylor sqrt(const lx_itaylor& x, int n)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    if (0<=x.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, sqrt(x,n): 0 in interval" 
                << std::endl;
		exit(1);
    };
    erg.tayl[0] = sqrt(x.tayl[0],n); // element No. 0
    for (int k=1; k<=order; k++)
    {
		erg.tayl[k] = 0;
		for (int j=0; j<=k-1; j++)
	   	erg.tayl[k] += (lx_interval(0,l_interval(k-j))/real(n)-interval(j))
		           * erg.tayl[j] * x.tayl[k-j];
		erg.tayl[k] /= (interval(k)*x.tayl[0]);
    }
    return erg;
}

//-----------------------------------------------------------------------

// sqrt(1+x^2)
lx_itaylor sqrt1px2(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg(order);
    const real c = 500.0;

    if (Inf(x.tayl[0]) > c) erg = x*sqrt(real(1)+real(1)/sqr(x));
    else 
		 if (Sup(x.tayl[0]) < -c) 
			 erg = -x*sqrt(real(1)+real(1)/sqr(x)); 
    	 else erg = sqrt(real(1)+sqr(x));
    return erg;
}

//-----------------------------------------------------------------------

// sqrt(1+x)-1
lx_itaylor sqrtp1m1(const lx_itaylor& x)
{
    int order(get_order(x)),m;
    lx_itaylor erg(order);
    lx_interval h,Ne;

    if (Inf(x.tayl[0])<=-1)
    {
		std::cerr << "Error in lx_itaylor, sqrtp1m1: wrong argument" 
                << std::endl;
		exit(1);
    };

    erg.tayl[0] = sqrtp1m1(x.tayl[0]);
    Ne = real(1.0)+erg.tayl[0];
    times2pown(Ne,1);

    for (int k=1; k<=order; k++)
    {
		m = (k+1)/2;
		h = lx_interval(0,l_interval(0));
		for (int j=1; j<=m-1; j++)
	    	h = h + erg.tayl[j]*erg.tayl[k-j];
		times2pown(h,1);
		erg.tayl[k] = h;
		if (k%2==0) 
			erg.tayl[k] += sqr(erg.tayl[m]);
		erg.tayl[k] = (x.tayl[k]-erg.tayl[k]) / Ne; 
    }
    return erg;
} // sqrtp1m1

//-----------------------------------------------------------------------

// sqrt(1-x^2):
lx_itaylor sqrt1mx2(const lx_itaylor& x)
{
    lx_interval sum,h;
    int order(get_order(x)),m;
    lx_itaylor erg(order), g(order);

    if (Inf(x.tayl[0])<=-1 || Sup(x.tayl[0])>=1)
    {
		std::cerr << "Error in lx_itaylor, sqrt1mx2: wrong argument" 
                << std::endl;
		exit(1);
    };
    erg.tayl[0]=sqrt1mx2(x.tayl[0]);
    h = real(-1)/erg.tayl[0];
    times2pown(h,-1);
    g = sqr(x);
    for (int k=1; k<=order; k++)
    {
		m = (k+1)/2;
		sum = lx_interval(0,l_interval(0));
		for (int j=1; j<=m-1; j++) 
	    	sum = sum + erg.tayl[j]*erg.tayl[k-j];
		times2pown(sum,1);  // Multiplication with 2
		erg.tayl[k] = sum;
		if (k%2==0) 
			erg.tayl[k] += sqr(erg.tayl[m]); // k even 
		erg.tayl[k] = (g.tayl[k]+erg.tayl[k])*h;
    }
    return erg;
}

//-----------------------------------------------------------------------

// sqrt(x^2-1):
lx_itaylor sqrtx2m1(const lx_itaylor& x)
{
    const real c = 30.0; 
    lx_interval sum,h;
    int order(get_order(x)),m;
    lx_itaylor erg(order), g(order);

	 if (Disjoint( x.tayl[0], lx_interval(0,interval(-1,1)) )==0)
    {
		std::cerr << "Error in lx_itaylor, sqrtx2m1: wrong argument" 
                << std::endl;
		exit(1);
    };

    if (Inf(x.tayl[0])>c) 
		 erg = x*sqrt1mx2(real(1)/x);
    else 
		 if (Sup(x.tayl[0])<-c) 
			 erg = -x*sqrt1mx2(real(1)/x);
       else 
		 {
			erg.tayl[0]=sqrtx2m1(x.tayl[0]); 
			g = sqr(x);
			h = real(1)/erg.tayl[0];
			times2pown(h,-1);
			for (int k=1; k<=order; k++)
			{
	    		m = (k+1)/2;
	    		sum = 0;
	    		for (int j=1; j<=m-1; j++) 
					sum = sum + erg.tayl[j]*erg.tayl[k-j];
	    		times2pown(sum,1);  // Multiplication with 2
	   		erg.tayl[k] = sum;
	    		if (k%2==0) 
					erg.tayl[k] += sqr(erg.tayl[m]); // k even 
	    		erg.tayl[k] = (g.tayl[k]-erg.tayl[k])*h;
			}
    	 }
    return erg;
}

//-----------------------------------------------------------------------

// power-function
lx_itaylor pow(const lx_itaylor& x, const lx_interval& alpha)
{
    int order(get_order(x));
    lx_itaylor erg(order);

    if (0<=x.tayl[0])
    {
		std::cerr << "Error in lx_itaylor, pow(x,a): 0 in interval x" 
                << std::endl;
		exit(1);
    };

    erg.tayl[0] = pow(x.tayl[0],alpha); // element No. 0
    for (int k=1; k<=order; k++)
    {
		erg.tayl[k] = 0;
		for (int j=0; j<=k-1; j++)
	      erg.tayl[k] += (interval(k-j)*alpha-interval(j))
		           * erg.tayl[j] * x.tayl[k-j];
		erg.tayl[k] /= (interval(k)*x.tayl[0]);
    }
    return erg;
}
/*
// power-function
lx_itaylor power(const lx_itaylor& x, int n)
{
	int order(get_order(x));
	lx_itaylor erg(order);

	erg.tayl[0] = power(x.tayl[0],n); // element No. 0
	if (order>0) // First derivative without division!
	{
		erg.tayl[1] = n*power(x.tayl[0],n-1)*x.tayl[1];
	}
	for (int k=2; k<=order; k++) // 2.,3., ... derivatives
	{  
		erg.tayl[k] = 0;
		for (int j=0; j<=k-1; j++)
			erg.tayl[k] += (interval(k-j)*n-interval(j))
					* erg.tayl[j] * x.tayl[k-j];
		erg.tayl[k] /= (interval(k)*x.tayl[0]);
	}
	return erg;
}
*/

lx_itaylor power(const lx_itaylor& x, int n)
{
	int order(get_order(x));
	lx_itaylor y(order), neu(order);
   bool neg = false;
   long int zhi = 2;
   if (n == 0)
		y = 1;
	else 
	{
		if (n == 1)
			y = x;
		else 
			if (n == 2)
				y = sqr(x);
				else 
				{
					if (n < 0) 
					{
						neg = true;
						n = -n;
					}
               // Initialisierung
					if (n%2)  
						y = x;
					else      
						y = 1;  // Praezision wird bei 1 Mult. auf
                            // aktuellen Wert gesetzt;
               // Berechnugn durch binaere Darstellung der n
					neu = sqr(x);   // neu = x*x;
					do 
					{
						if ((n/zhi)%2)  
							y = y * neu;
						zhi += zhi;
						if (zhi <= n)  // letzte Mult. entfaellt --> schneller
							neu = neu * neu;
					} while (zhi <= n);

					if (neg) 
						y = 1/(y);
				}
	}
	return y;
}

//-----------------------------------------------------------------------

// Exponential-function
lx_itaylor exp(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg(order);

    erg.tayl[0] = exp(x.tayl[0]); // element No. 0; function value
    for(int k=1; k<=order; k++)
    {
		erg.tayl[k] = 0;
		for(int j=0; j<=k-1; j++)
	    	erg.tayl[k] += interval(k-j)*erg.tayl[j]*x.tayl[k-j]; 
		erg.tayl[k] /= interval(k);
    }
    return erg; 
}

lx_itaylor exp2(const lx_itaylor& x)
{
	int order(get_order(x));
	lx_itaylor erg(order);

	erg.tayl[0] = exp2(x.tayl[0]); // element No. 0; function value
	for(int k=1; k<=order; k++)
	{
		erg.tayl[k] = 0;
		for(int j=0; j<=k-1; j++)
			erg.tayl[k] += interval(k-j)*erg.tayl[j]*x.tayl[k-j];
		erg.tayl[k] = ( erg.tayl[k]*Ln2_lx_interval() )/interval(k);
	}
	return erg; 
}

lx_itaylor exp10(const lx_itaylor& x)
{
	int order(get_order(x));
	lx_itaylor erg(order);

	erg.tayl[0] = exp10(x.tayl[0]); // element No. 0; function value
	for(int k=1; k<=order; k++)
	{
		erg.tayl[k] = 0;
		for(int j=0; j<=k-1; j++)
			erg.tayl[k] += interval(k-j)*erg.tayl[j]*x.tayl[k-j];
		erg.tayl[k] = ( erg.tayl[k]*Ln10_lx_interval() )/interval(k);
	}
	return erg; 
}

//-----------------------------------------------------------------------

// exp(x)-1;

lx_itaylor expm1(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg(order);

    erg.tayl[0] = exp(x.tayl[0]); // element No. 0; function value
    for(int k=1; k<=order; k++)
    {
		erg.tayl[k] = 0;
		for(int j=0; j<=k-1; j++)
	    	erg.tayl[k] += interval(k-j)*erg.tayl[j]*x.tayl[k-j]; 
		erg.tayl[k] /= interval(k);
    }
    erg.tayl[0] = expm1(x.tayl[0]);
    return erg; 
}


//-----------------------------------------------------------------------

// Help function
void f_g_u(lx_itaylor& f, const lx_itaylor& g, const lx_itaylor& u, int nb_function)
{
	int order1=get_order(f);
 	int order2=get_order(g);
 	int order3=get_order(u);

 // The following errors should be caught before 
 // but for security here again:
 	if(order1 != order2) 
   {
   	std::cerr << "Error1 in f_g_u: different orders " 
                << std::endl;
      exit(1);
   };

 	if(order3 != order2) 
   {
   	std::cerr << "Error2 in f_g_u: different orders " << std::endl;
   	exit(1);
  	};

 	if(0 <= g.tayl[0]) 
  	{
   	std::cerr << "Error in f_g_u : wrong argument " << std::endl;
   	exit(1);
  	}; 

 	switch(nb_function) // element No. 0
   {
    case _i_ln:f.tayl[0]=ln(u.tayl[0]); break;

    case _i_tan:f.tayl[0]=tan(u.tayl[0]); break;
    case _i_cot:f.tayl[0]=cot(u.tayl[0]); break;

    case _i_asin:f.tayl[0]=asin(u.tayl[0]); break;
    case _i_acos:f.tayl[0]=acos(u.tayl[0]); break;
    case _i_atan:f.tayl[0]=atan(u.tayl[0]); break;
    case _i_acot:f.tayl[0]=acot(u.tayl[0]); break;

    case _i_tanh:f.tayl[0]=tanh(u.tayl[0]); break;
    case _i_coth:f.tayl[0]=coth(u.tayl[0]); break; 

    case _i_asinh:f.tayl[0]=asinh(u.tayl[0]); break;
    case _i_acosh:f.tayl[0]=acosh(u.tayl[0]); break;
    case _i_atanh:f.tayl[0]=atanh(u.tayl[0]); break;

    case _i_acoth:f.tayl[0]=acoth(u.tayl[0]); break;

   }
 
 // remaining elements:
 lx_interval sum; 
 for(int j=1; j<=Ub(f.tayl); j++) 
 {
     sum = lx_interval(0,l_interval(0.0));
     for(int i=1; i<=j-1; i++)
	  	  sum += interval(i)*f.tayl[i]*g.tayl[j-i];
     f.tayl[j] = (u.tayl[j]-sum/interval(j)) / g.tayl[0];
 }
} // f_g_u

//-----------------------------------------------------------------------

// Logarithm-function
lx_itaylor ln(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);

    f_g_u(f,x,x,_i_ln);
  
    return f;
}

lx_itaylor log2(const lx_itaylor& x)
{
	int order=get_order(x);
	lx_itaylor res(order);

	res = ln(x)/Ln2_lx_interval();
  
	return res;
}

lx_itaylor log10(const lx_itaylor& x)
{
	int order=get_order(x);
	lx_itaylor res(order);

	res = ln(x)/Ln10_lx_interval();
  
	return res;
}

//-----------------------------------------------------------------------

// ln(1+x)
lx_itaylor lnp1(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg(order), g(order);

    g = interval(1) + x;
    erg.tayl[0] = lnp1(x.tayl[0]);
    for (int k=1; k<=order; k++)
    {
	 	erg.tayl[k] = 0;
		for (int j=1; j<=k-1; j++)
	     erg.tayl[k] += interval(j) * erg.tayl[j] * g.tayl[k-j];
		erg.tayl[k] = (x.tayl[k]-erg.tayl[k]/interval(k)) / g.tayl[0];
    }
    return erg;
}

//-----------------------------------------------------------------------

// Sinus-function
lx_itaylor sin(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg1(order);   // sin
    lx_itaylor erg2(order);   // cos
    lx_interval s1,s2;

    erg1.tayl[0]=sin(x.tayl[0]); // Element No. 0:  erg1 (sin)
    erg2.tayl[0]=cos(x.tayl[0]); // Element No. 0:  erg2 (cos)
    // remainig elements: 
    for(int j=1; j<=Ub(x.tayl); j++) 
    {
		s1=s2=lx_interval(0,l_interval(0));
		for(int i=0; i<=j-1; i++)
		{
	    	s1 += interval(j-i) * erg2.tayl[i] * x.tayl[j-i];
	    	s2 += interval(j-i) * erg1.tayl[i] * x.tayl[j-i];
		}
	   erg1.tayl[j]= s1/interval(j);
		erg2.tayl[j]= -s2/interval(j);
    }
    return erg1; 
}

//-----------------------------------------------------------------------

// Cosinus-function
lx_itaylor cos(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg1(order);   // sin
    lx_itaylor erg2(order);   // cos
    lx_interval s1,s2;

    erg1.tayl[0]=sin(x.tayl[0]); // Element No. 0:  erg1 (sin)
    erg2.tayl[0]=cos(x.tayl[0]); // Element No. 0:  erg2 (cos)

    // remainig elements: 
    for(int j=1; j<=Ub(x.tayl); j++) 
    {
		s1=s2=lx_interval(0,l_interval(0.0));
		for(int i=0; i<=j-1; i++)
		{
	    	s1 += interval(j-i) * erg2.tayl[i] * x.tayl[j-i];
	    	s2 += interval(j-i) * erg1.tayl[i] * x.tayl[j-i];
		}
		erg1.tayl[j]= s1/interval(j);
		erg2.tayl[j]= -s2/interval(j);
    }
 return erg2; 
}

//-----------------------------------------------------------------------

// Tangens-function
lx_itaylor tan(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = sqr(cos(x));

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, tan : wrong argument" << std::endl;
		exit(1);
    };
 
    f_g_u(f,g,x,_i_tan);
  
    return f;
}

//-----------------------------------------------------------------------

// Cotangens-function
lx_itaylor cot(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = -sqr(sin(x));

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, cot : wrong argument" << std::endl;
	   exit(1);
    };
 
    f_g_u(f,g,x,_i_cot);
  
    return f;
}

//-----------------------------------------------------------------------

// Sinushyperbolicus-function
lx_itaylor sinh(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg1(order);  // sinh
    lx_itaylor erg2(order);  // cosh
    lx_interval s1,s2;

    erg1.tayl[0]=sinh(x.tayl[0]); // element No. 0:  erg1 (sinh)
    erg2.tayl[0]=cosh(x.tayl[0]); // element No. 0:  erg2 (cosh)

    // remainig elements: 
    for(int j=1; j<=Ub(x.tayl); j++) 
    {
		s1=s2=interval(0);
		for(int i=0; i<=j-1; i++)
		{
	    	s1 += interval(j-i) * erg2.tayl[i] * x.tayl[j-i];
	    	s2 += interval(j-i) * erg1.tayl[i] * x.tayl[j-i];
		}
		erg1.tayl[j]= s1/interval(j);
		erg2.tayl[j]= s2/interval(j);
    }
    return erg1; 
}

//-----------------------------------------------------------------------

// Cosinushyperbolicus-function
lx_itaylor cosh(const lx_itaylor& x)
{
    int order(get_order(x));
    lx_itaylor erg1(order); // sinh
    lx_itaylor erg2(order); // cosh
    lx_interval s1,s2;

    erg1.tayl[0]=sinh(x.tayl[0]); // element No. 0:  erg1 (sinh)
    erg2.tayl[0]=cosh(x.tayl[0]); // element No. 0:  erg2 (cosh)

    // remaining elements: 
    for(int j=1; j<=Ub(x.tayl); j++) 
    {
		s1=s2=interval(0);
		for(int i=0; i<=j-1; i++)
		{
	    	s1 += interval(j-i) * erg2.tayl[i] * x.tayl[j-i];
	    	s2 += interval(j-i) * erg1.tayl[i] * x.tayl[j-i];
		}
		erg1.tayl[j]= s1/interval(j);
		erg2.tayl[j]= s2/interval(j);
    }
    return erg2; 
}

//-----------------------------------------------------------------------

//Tangenshyperbolicus-function
lx_itaylor tanh(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = sqr(cosh(x)); 
 
    f_g_u(f,g,x,_i_tanh);
  
    return f;
}

//-----------------------------------------------------------------------
 
// Cotangenshyperbolicus-function
lx_itaylor coth(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = -sqr(sinh(x));

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, coth : wrong argument " 
                << std::endl;
		exit(1);
    };  
 
    f_g_u(f,g,x,_i_coth);
  
    return f;
}

//-----------------------------------------------------------------------

//Arcsinusfunktion
lx_itaylor asin(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = sqrt1mx2(x);

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, asin : wrong argument " 
                << std::endl;
		exit(1);
    };
 
    f_g_u(f,g,x,_i_asin);
  
    return f;
}

//-----------------------------------------------------------------------

//Arccosinusfunktion
lx_itaylor acos(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = -sqrt1mx2(x);

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, acos : wrong argument " 
                << std::endl;
		exit(1);
    };
 
    f_g_u(f,g,x,_i_acos);
  
    return f;
}

//-----------------------------------------------------------------------

//Arctan-function
lx_itaylor atan(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = interval(1.0)+sqr(x);

    f_g_u(f,g,x,_i_atan);
  
    return f;
}

//-----------------------------------------------------------------------

//Arccotan-function
lx_itaylor acot(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = -(interval(1.0)+sqr(x));

    f_g_u(f,g,x,_i_acot);
  
    return f;
}

//-----------------------------------------------------------------------

//Areasinh-function
lx_itaylor asinh(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g=sqrt1px2(x);

    f_g_u(f,g,x,_i_asinh);
  
    return f;
}

//-----------------------------------------------------------------------

//Areacosh-function
lx_itaylor acosh(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g=sqrtx2m1(x);

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, acosh : wrong argument " 
                << std::endl;
		exit(1);
    };
 
    f_g_u(f,g,x,_i_acosh);
  
    return f;
}

//-----------------------------------------------------------------------

//Areatanh-function
lx_itaylor atanh(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g=interval(1.0)-sqr(x);

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, atanh : wrong argument " 
                << std::endl;
		exit(1);
    };
 
    f_g_u(f,g,x,_i_atanh);
  
    return f;
}

//-----------------------------------------------------------------------

//Areacotanh-function
lx_itaylor acoth(const lx_itaylor& x)
{
    int order=get_order(x);
    lx_itaylor f(order);
    lx_itaylor g(order);

    g = interval(1.0)-sqr(x);

    if(0 <= g.tayl[0]) 
    {
		std::cerr << "Error in lx_itaylor, acoth : wrong argument " 
                << std::endl;
		exit(1);
    };

    f_g_u(f,g,x,_i_acoth);
  
    return f;
}

void fEval ( df_FctPtr f, lx_interval x, lx_interval& fx)
{
	lx_itaylor fz;
	lx_itaylor z(0,x);
	fz = f(z);
	fx = get_j_derivative(fz,0);
}

void dfEval ( df_FctPtr f, lx_interval x, lx_interval& fx,
				                              lx_interval& dfx)
{
	lx_itaylor fz;
	lx_itaylor z(1,x);
	fz = f(z);
	fx  = get_j_derivative(fz,0); // Function value   f (x)
	dfx = get_j_derivative(fz,1); // First derivative f'(x)
}

} // end namespace taylor
